==> hyperdiv-0.0.0-py3-none-any.whl: hyperdiv/core/element.py <==
from collections import namedtuple
from functools import cache
from .context import Context


def to_caml_case(name):
    parts = name.split("_")
    return "".join([parts[0], *[part.capitalize() for part in parts[1:]]])


Prop = namedtuple(
    "Prop",
    ["value", "is_event_prop", "parse"],
    defaults=[None, False, lambda x: x],
)


class RemoveProp(Prop):
    pass


def parse_option(*values):
    def parser(value):
        if value in values:
            return value
        raise ValueError(f"Invalid value {value}.")

    return parser


class DynamicProp:
    def __init__(
        self,
        parent_element,
        name,
        value=None,
        is_event_prop=False,
        parse=lambda x: x,
    ):
        self.parent_element = parent_element
        self.name = name
        self.value = value
        self.is_event_prop = is_event_prop
        self._parse = parse
        self._cached_parse = None
        self.mutated = False

        self._initial_value = value

    @property
    def context(self):
        return self.parent_element._context

    def parse(self, value):
        # Parses the prop if the value has changed.
        # Otherwise it returns the cached parsed value.
        # Assumes that values can be compared using == and !=
        if not self._cached_parse or value != self._cached_parse[0]:
            self._cached_parse = (value, self._parse(value))

        return self._cached_parse[1]

    def init(self, value):
        if not self.mutated:
            if callable(value):
                self.parse(value(self.parent_element))
            else:
                self.parse(value)
            self.value = value

    def get_value(self):
        in_user_code = self.context.in_user_code()

        if in_user_code and self.is_event_prop:
            # If we're an event prop like `clicked`, etc., we record
            # user-code accesses into the read set.
            self.context.read_event_props.add(self)

        if callable(self.value):
            return self.value(self.parent_element)
        return self.value

    def set_value(self, value):
        in_user_code = self.context.in_user_code()

        if callable(self.value):
            self.parse(self.value(self.parent_element))
        else:
            self.parse(value)

        if in_user_code and self.is_event_prop:
            # Event props are read-only from user code. Cannot write
            # to `clicked`, etc.
            raise Exception(f"Cannot write to even prop '{self.name}' from user code.")

        if self.is_event_prop:
            # We track mutated event props (clicked, etc.) so we can
            # reset them at the end of the frame.
            self.context.mutated_event_props.add(self)

        if in_user_code and value != self.value:
            self.context.frame_had_mutations.set()

        self.mutated = True
        self.value = value

    def clear(self):
        self.mutated = False

    def reset(self):
        self.set_value(self._initial_value)


def get_element_name(element_class):
    try:
        return element_class.name
    except AttributeError:
        # If a name is not set on the class, fall back to the class's
        # name.
        return element_class.__name__


class ElementCacheMetaclass(type):
    def __call__(cls, *args, **kwargs):
        context = Context.current()

        if cls.is_singleton:
            # TODO: A hash of the name would avoid possible collisions
            # with other user-supplied keys.
            key = get_element_name(cls)
        else:
            key = context.create_element_key(kwargs.pop("key", None))

        comp = context.element_state.get_element(key)
        if not comp:
            comp = super(ElementCacheMetaclass, cls).__call__(*args, key=key, **kwargs)
            context.element_state.set_element(comp)

        comp._init(*args, **kwargs)

        return comp


@cache
def collect_static_props(klass):
    """Walk up the class hierarchy, starting at klass, and collect
    all the `Prop`s into a flat dict.
    """
    props = dict()

    def collect_props(klass):
        """Post-order traversal makes sure that when we update the `props`
        dict, a prop in a subclass will override a prop with the
        same name in a superclass.
        """
        for b in klass.__bases__:
            if b != Element and issubclass(b, Element):
                collect_props(b)
        for k, v in vars(klass).items():
            if isinstance(v, RemoveProp):
                props.pop(k, None)
            elif isinstance(v, Prop):
                props[k] = v

    collect_props(klass)
    return props


class PropAccessor:
    def __init__(self, props):
        self._props = props

    def __getattribute__(self, key):
        try:
            is_prop = key != "_props" and key in self._props
        except AttributeError:
            is_prop = False

        if is_prop:
            return self._props[key].get_value()
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        try:
            is_prop = key != "_props" and key in self._props
        except AttributeError:
            is_prop = False

        if is_prop:
            return self._props[key].set_value(value)
        else:
            return super().__setattr__(key, value)


class Element(PropAccessor, metaclass=ElementCacheMetaclass):
    # Defaults
    is_collected = True
    is_container = False
    is_singleton = False

    def __init__(self, *args, **kwargs):
        self._context = Context.current()
        self.key = kwargs.pop("key")

        self.name = get_element_name(self.__class__)
        if self.is_container:
            self.children = []

        self.was_collected = False

        def reset_collected():
            self.was_collected = False

        self._context.bus.on("frame_start", reset_collected)

        static_props = collect_static_props(self.__class__)

        dynamic_props = dict()
        by_id = dict()  # Account for aliasing
        for name, static_prop in static_props.items():
            prop_id = id(static_prop)
            if prop_id not in by_id:
                by_id[prop_id] = DynamicProp(self, name, *static_prop)
            dynamic_props[name] = by_id[prop_id]

        super().__init__(dynamic_props)

    #
    # Methods that are designed to be overridden by element
    # writers. They all take as input the args that the user supplies
    # when invoking the element.
    #

    def args_to_props(self, *args, **kwargs):
        """Given the args/kwargs passed by the user to the element (minus
        the key kwarg), return a dict from prop names to prop values
        that will be used to initialize the props.

        By default, args is ignored and kwargs is assumed to map valid
        prop names to values.
        """
        return kwargs

    def on_render(*args, **kwargs):
        """Override this to do work after the element has been fully set
        up. This method is called on each frame in which the element
        is rendered.
        """
        pass

    #
    # The rest of these methods should in general not be overridden:
    #

    def _init_clear(self):
        if self.is_singleton and (not self.is_collected or self.was_collected):
            return
        self.clear()

    def _init_collect(self):
        if self.is_collected:
            if self.is_singleton and (not self.is_collected or self.was_collected):
                return
            self._context.collector_stack.collect(self)
            self.was_collected = True

    def _init_props(self, *args, **kwargs):
        init_values = self.args_to_props(*args, **kwargs)

        for prop_name, prop_value in init_values.items():
            prop = self._props.get(prop_name)
            if not prop:
                raise Exception(f'Unexpected property "{prop_name}" on "{self.name}".')
            prop.init(prop_value)

    def _init(self, *args, **kwargs):
        self._init_clear()
        self._init_collect()
        self._init_props(*args, **kwargs)

        self.on_render(*args, **kwargs)

    def mutated(self, propname):
        return self._props[propname].mutated

    def clear_and_reset_prop(self, propname):
        self._props[propname].clear()
        self._props[propname].reset()

    def clear(self):
        if self.is_container:
            self.children.clear()

    def to_dom(self):
        dom = dict(
            _dom=True,
            name=self.name,
            key=self.key,
            **{
                to_caml_case(prop_name): prop.parse(prop.get_value())
                for prop_name, prop in self._props.items()
                # We don't send event props to the UI.
                if not prop.is_event_prop
            },
        )
        if self.is_container:
            dom["children"] = [element.to_dom() for element in self.children]
        return dom

    def __enter__(self):
        if not self.is_container:
            raise Exception(
                f"A '{self.name}' cannot be used in a `with` context "
                "because it's not a container."
            )
        self._context.collector_stack.push(self.children)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context.collector_stack.pop()

    def __str__(self):
        props = " ".join(
            [f"{prop.name}={prop.get_value()}" for prop in self._props.values()]
        )
        return f"<{self.name} {props}>"

    def __repr__(self):
        return str(self)

==> hyperdiv-0.0.1-py3-none-any.whl: hyperdiv/core/element.py <==
from collections import namedtuple
from functools import cache
from .context import Context


def to_caml_case(name):
    parts = name.split("_")
    return "".join([parts[0], *[part.capitalize() for part in parts[1:]]])


Prop = namedtuple(
    "Prop",
    ["value", "is_event_prop", "parse"],
    defaults=[None, False, lambda x: x],
)


class RemoveProp(Prop):
    pass


def parse_option(*values):
    def parser(value):
        if value in values:
            return value
        raise ValueError(f"Invalid value {value}.")

    return parser


class DynamicProp:
    def __init__(
        self,
        parent_element,
        name,
        value=None,
        is_event_prop=False,
        parse=lambda x: x,
    ):
        self.parent_element = parent_element
        self.name = name
        self.value = value
        self.is_event_prop = is_event_prop
        self._parse = parse
        self._cached_parse = None
        self.mutated = False

        self._initial_value = value

    @property
    def context(self):
        return self.parent_element._context

    def parse(self, value):
        # Parses the prop if the value has changed.
        # Otherwise it returns the cached parsed value.
        # Assumes that values can be compared using == and !=
        if not self._cached_parse or value != self._cached_parse[0]:
            self._cached_parse = (value, self._parse(value))

        return self._cached_parse[1]

    def init(self, value):
        if not self.mutated:
            if callable(value):
                self.parse(value(self.parent_element))
            else:
                self.parse(value)
            self.value = value

    def get_value(self):
        in_user_code = self.context.in_user_code()

        if in_user_code and self.is_event_prop:
            # If we're an event prop like `clicked`, etc., we record
            # user-code accesses into the read set.
            self.context.read_event_props.add(self)

        if callable(self.value):
            return self.value(self.parent_element)
        return self.value

    def _register_change(self, new_value):
        in_user_code = self.context.in_user_code()
        if in_user_code and new_value != self.value:
            self.context.frame_had_changes.set()

    def set_value(self, value):
        in_user_code = self.context.in_user_code()

        if callable(self.value):
            self.parse(self.value(self.parent_element))
        else:
            self.parse(value)

        if in_user_code and self.is_event_prop:
            # Event props are read-only from user code. Cannot write
            # to `clicked`, etc.
            raise Exception(f"Cannot write to even prop '{self.name}' from user code.")

        if self.is_event_prop:
            # We track mutated event props (clicked, etc.) so we can
            # reset them at the end of the frame.
            self.context.mutated_event_props.add(self)

        self._register_change(value)

        self.mutated = True
        self.value = value

    def reset(self):
        self._register_change(self._initial_value)
        self.value = self._initial_value
        self.mutated = False


def get_element_name(element_class):
    try:
        return element_class.name
    except AttributeError:
        # If a name is not set on the class, fall back to the class's
        # name.
        return element_class.__name__


class ElementCacheMetaclass(type):
    def __call__(cls, *args, **kwargs):
        context = Context.current()

        if cls.is_singleton:
            # TODO: A hash of the name would avoid possible collisions
            # with other user-supplied keys.
            key = get_element_name(cls)
        else:
            key = context.create_element_key(kwargs.pop("key", None))

        comp = context.element_state.get_element(key)
        if not comp:
            comp = super(ElementCacheMetaclass, cls).__call__(*args, key=key, **kwargs)
            context.element_state.set_element(comp)

        comp._init(*args, **kwargs)

        return comp


@cache
def collect_static_props(klass):
    """Walk up the class hierarchy, starting at klass, and collect
    all the `Prop`s into a flat dict.
    """
    props = dict()

    def collect_props(klass):
        """Post-order traversal makes sure that when we update the `props`
        dict, a prop in a subclass will override a prop with the
        same name in a superclass.
        """
        for b in klass.__bases__:
            if b != Element and issubclass(b, Element):
                collect_props(b)
        for k, v in vars(klass).items():
            if isinstance(v, RemoveProp):
                props.pop(k, None)
            elif isinstance(v, Prop):
                props[k] = v

    collect_props(klass)
    return props


class PropAccessor:
    def __init__(self, props):
        self._props = props

    def __getattribute__(self, key):
        try:
            is_prop = key != "_props" and key in self._props
        except AttributeError:
            is_prop = False

        if is_prop:
            return self._props[key].get_value()
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        try:
            is_prop = key != "_props" and key in self._props
        except AttributeError:
            is_prop = False

        if is_prop:
            return self._props[key].set_value(value)
        else:
            return super().__setattr__(key, value)


class Element(PropAccessor, metaclass=ElementCacheMetaclass):
    # Defaults
    is_collected = True
    is_container = False
    is_singleton = False

    def __init__(self, *args, **kwargs):
        self._context = Context.current()
        self.key = kwargs.pop("key")

        self.name = get_element_name(self.__class__)
        if self.is_container:
            self.children = []

        self.was_collected = False

        def reset_collected():
            self.was_collected = False

        self._context.bus.on("frame_start", reset_collected)

        static_props = collect_static_props(self.__class__)

        dynamic_props = dict()
        by_id = dict()  # Account for aliasing
        for name, static_prop in static_props.items():
            prop_id = id(static_prop)
            if prop_id not in by_id:
                by_id[prop_id] = DynamicProp(self, name, *static_prop)
            dynamic_props[name] = by_id[prop_id]

        super().__init__(dynamic_props)

    #
    # Methods that are designed to be overridden by element
    # writers. They all take as input the args that the user supplies
    # when invoking the element.
    #

    def args_to_props(self, *args, **kwargs):
        """Given the args/kwargs passed by the user to the element (minus
        the key kwarg), return a dict from prop names to prop values
        that will be used to initialize the props.

        By default, args is ignored and kwargs is assumed to map valid
        prop names to values.
        """
        return kwargs

    def on_render(*args, **kwargs):
        """Override this to do work after the element has been fully set
        up. This method is called on each frame in which the element
        is rendered.
        """
        pass

    #
    # The rest of these methods should in general not be overridden:
    #

    def _init_clear(self):
        if self.is_singleton and (not self.is_collected or self.was_collected):
            return
        self.clear()

    def _init_collect(self):
        if self.is_collected:
            if self.is_singleton and (not self.is_collected or self.was_collected):
                return
            self._context.collector_stack.collect(self)
            self.was_collected = True

    def _init_props(self, *args, **kwargs):
        init_values = self.args_to_props(*args, **kwargs)

        for prop_name, prop_value in init_values.items():
            prop = self._props.get(prop_name)
            if not prop:
                raise Exception(f'Unexpected property "{prop_name}" on "{self.name}".')
            prop.init(prop_value)

    def _init(self, *args, **kwargs):
        self._init_clear()
        self._init_collect()
        self._init_props(*args, **kwargs)

        self.on_render(*args, **kwargs)

    def mutated(self, propname=None):
        """If `propname` is given, return True if that prop is mutated. If
        `propname` is not given, return True if *any* prop is mutated.
        """

        if propname:
            prop = self._props.get(propname)
            return prop and prop.mutated
        else:
            for prop in self._props.values():
                if prop.mutated:
                    return True
            return False

    def reset_prop(self, propname):
        self._props[propname].reset()

    def clear(self):
        if self.is_container:
            self.children.clear()

    def to_dom(self):
        dom = dict(
            _dom=True,
            name=self.name,
            key=self.key,
            **{
                to_caml_case(prop_name): prop.parse(prop.get_value())
                for prop_name, prop in self._props.items()
                # We don't send event props to the UI.
                if not prop.is_event_prop
            },
        )
        if self.is_container:
            dom["children"] = [element.to_dom() for element in self.children]
        return dom

    def __enter__(self):
        if not self.is_container:
            raise Exception(
                f"A '{self.name}' cannot be used in a `with` context "
                "because it's not a container."
            )
        self._context.collector_stack.push(self.children)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context.collector_stack.pop()

    def __str__(self):
        props = " ".join(
            [f"{prop.name}={prop.get_value()}" for prop in self._props.values()]
        )
        return f"<{self.name} {props}>"

    def __repr__(self):
        return str(self)
